// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#ifndef OUTPUTREVISION_H
#define OUTPUTREVISION_H

#include <QtCore/qglobal.h> // -- QtScxml

QT_BEGIN_NAMESPACE // -- QtScxml

// if the output revision changes, you MUST change it in qtmetamacros.h too
enum { mocOutputRevision = 69 };          // moc format output revision

QT_END_NAMESPACE // -- QtScxml

#endif // OUTPUTREVISION_H
