// Copyright (C) 2021 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant

#include "qscxmldatamodelplugin_p.h"

QT_BEGIN_NAMESPACE

QScxmlDataModel *QScxmlDataModelPlugin::createScxmlDataModel() const
{
    return nullptr;
}

QT_END_NAMESPACE
