// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#ifndef QSCXMLC_H
#define QSCXMLC_H

#include <QtCore/qglobal.h>

QT_BEGIN_NAMESPACE

int run(const QStringList &arguments);

QT_END_NAMESPACE

#endif // QSCXMLC_H
