// Copyright (C) 2016 Klaralvdalens Datakonsult AB (KDAB).
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include "connection.h"

int main(int argc, char **argv)
{
  Q_UNUSED(argc);
  Q_UNUSED(argv);

  Connection connection;
  Q_UNUSED(connection);
  return 0;
}
